/*	$Id: dynload.c 2007/08/21 19:07:57 mallet $ */

/*
 * Copyright (c) 2000-2007 LAAS/CNRS                  --  Tue May  9 2000
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
#ident "$Id: dynload.c 2007/08/21 19:07:57 mallet $"

#ifdef VXWORKS
#include <vxWorks.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <limits.h>

#include "dynload.h"

/*
 * insituLoadPackage ----------------------------------------------------
 *
 * Dynamically loads a package (a set of functions that compute matrices)
 * and returns an InsituPkg struct.
 * The returned struct is malloc'ed.
 */

InsituPkg *
insituLoadPackage(char *path, char *name)
{
   char *file, *localpath;
   char module_name[PATH_MAX];
#ifdef VXWORKS
   int fd;
   static char objSuffix[] = ".o";
#else
   static char objSuffix[] = ".so";
#endif

   InsituPkg* pkg;

   pkg = malloc(sizeof(InsituPkg));
   if (pkg == NULL) { errno = ENOMEM; return NULL; }
   pkg->handle = NULL;
   pkg->next = NULL;

   localpath = malloc(strlen(path)+1);
   if (localpath == NULL) {
      free(pkg);
      errno = ENOMEM;
      return NULL;
   }
   strcpy(localpath, path);

   for(file = strtok(localpath, ":");
       file != NULL; file = strtok(NULL, ":")) {
      sprintf(module_name, "%s/%s%s", file, name, objSuffix);

#ifndef VXWORKS
      pkg->handle = dlopen(module_name, RTLD_LAZY);
      if (pkg->handle != NULL) break;
#else
      fd = open(module_name, O_RDONLY, 0);
      if (fd != -1) {
	 pkg->handle = loadModule(fd, LOAD_GLOBAL_SYMBOLS);
	 if (pkg->handle != NULL) {
	    close(fd);
	    break;
	 }
	 close(fd);
      }
#endif
   }
   free(localpath);

   if (pkg->handle == NULL) {
#ifndef VXWORKS
      fprintf(stderr, "insituLoadPackage: %s\n", dlerror());
      errno = EINVAL;
#else
      fprintf(stderr, "insituLoadPackage: %s\n", strerror(errno));
      fprintf(stderr, "insituLoadPackage: for %s\n", name);
#endif
      free(pkg);
      return NULL;
   }

   pkg->name = malloc(strlen(name)+1);
   if (pkg->name != NULL) strcpy(pkg->name, name);
   pkg->path = malloc(strlen(module_name)+1);
   if (pkg->path != NULL) strcpy(pkg->path, module_name);

   return pkg;
}


/*
 * insituGetSymbol ------------------------------------------------------
 *
 * Looks for a symbol "sym" in the package "pkg" and returns a pointer on
 * it or NULL if "sym" cannot be found.
 */

void *
insituGetSymbol(InsituPkg *pkg, char *sym)
{
   void *psym;
#ifdef VXWORKS
   SYM_TYPE symType;
   char sym_underscore[PATH_MAX];

   sprintf(sym_underscore, "_%s", sym);
#endif

   for(; pkg != NULL ; pkg = pkg->next) {
#ifndef VXWORKS
      psym = dlsym(pkg->handle, sym);
      if (psym != NULL) break;
#else
      if (symFindByName(sysSymTbl, sym,
			(char **)&psym, &symType) == OK)
	 break;
      if (symFindByName(sysSymTbl, sym_underscore,
			(char **)&psym, &symType) == OK)
	 break;
      psym = NULL;
#endif
   }

   if (psym == NULL) {
      errno = EINVAL;
      return NULL;
   }

   return psym;
}
