/*	$Id: geo.c 2007/08/21 18:30:43 mallet $ */

/*
 * Copyright (c) 2000-2007 LAAS/CNRS                  --  Fri May 12 2000
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
#ident "$Id: geo.c 2007/08/21 18:30:43 mallet $"

#ifdef VXWORKS
#include <vxWorks.h>
#include <semLib.h>
#endif

#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <stdlib.h>
#include <errno.h>

#include "insitu/lib.h"
#include "geo.h"
#include "graph.h"

/* Local functions */
static InsituGeom *	newGeom(ino_t ino, dev_t dev, char *path,
				char *file);

/* We keep a reference on every geometrical description that is read to
 * avoid multiple initializations with the same config file. Under Vxowrks,
 * we use a binary semaphore to pretect acces to this list (see
 * insituInitRobot). */

static InsituGeom *geoms = NULL;
#ifdef VXWORKS
static SEM_ID geomSemId = NULL;
#endif

/*
 * insituNewGeom --------------------------------------------------------
 *
 * Read a description file and returns the handle on it. If the file has
 * already been parsed, the function simply returns the previously created
 * handle so that it is shared between the different callers.
 *
 * This is a critical section since if it two tasks (or threads) want to
 * read the same file, one of them must wait for the other to complete in
 * order to actually share the geometrical description.
 * Under VxWorks, this can simply be done with a binary semaphore. Under
 * Unix, it is a bit weirder since that library can either be used in a
 * multi-threaded context or not. Most of the time, the Unix version won't
 * be accessed concurently (even in a multi-threaded context). So there is
 * no protection for this section under Unix, util a better solution is
 * found.
 */

InsituGeom *
insituNewGeom(char *geomFile)
{
   FILE *f;
   InsituGeom *g;
   char *file, *path, *localpath;
   char module_name[PATH_MAX];
   struct stat statBuf;
   extern void geomNewFile(FILE *f, char *file);
   extern int geomparse(void *geom);


#ifdef VXWORKS
  createSem:
   if (geomSemId == NULL) {
      geomSemId = semBCreate(SEM_Q_PRIORITY, SEM_FULL);
      if (geomSemId == NULL) return NULL;
   }

   if (semTake(geomSemId, WAIT_FOREVER) != OK) {
      /* The semaphore might have been deleted meanwhile because the
       * last geometrical model might have been destroyed. If so, we just
       * have to create the semaphore again. */
      if (geomSemId == NULL) goto createSem;
      return NULL;
   }
#endif

   if (geomFile[0] == '/') path = "/"; else path = STDPATH;
   localpath = malloc(strlen(path)+1);
   if (localpath == NULL) {
#ifdef VXWORKS
      semGive(geomSemId);
#endif
      errno = ENOMEM;
      return NULL;
   }
   strcpy(localpath, path);

   for(file = strtok(localpath, ":");
       file != NULL; file = strtok(NULL, ":")) {
      sprintf(module_name, "%s/%s", file, geomFile);
      if (!stat(module_name, &statBuf)) break;
   }

   free(localpath);

   if (file == NULL) {
#ifdef VXWORKS
      semGive(geomSemId);
#endif
      return NULL;
   }

   /* If that file has already been read, simply return the corresponding
    * struct */

   for(g = geoms; g != NULL; g = g->next)
      if (g->ino == statBuf.st_ino && g->dev == statBuf.st_dev) {
	 g->refCount++;
#ifdef VXWORKS
	 semGive(geomSemId);
#endif
	 return g;
      }

   /* The file is not known yet. Read it */
   g = newGeom(statBuf.st_ino, statBuf.st_dev, path, module_name);
   if (g == NULL) {
#ifdef VXWORKS
      semGive(geomSemId);
#endif
      return NULL;
   }

   f = fopen(g->file, "r");
   if (f == NULL) {
#ifdef VXWORKS
      semGive(geomSemId);
#endif
      return NULL;
   }
   geomNewFile(f, g->file);
   if (geomparse(g)) {
#ifdef VXWORKS
      semGive(geomSemId);
#endif
      return NULL;
   }
   fclose(f);
   if (insituCleanUpGeom(g)) {
#ifdef VXWORKS
      semGive(geomSemId);
#endif
      return NULL;
   }

   g->next = geoms;
   geoms = g;

#ifdef VXWORKS
   semGive(geomSemId);
#endif
   return g;
}


/*
 * insituWriteGeom ------------------------------------------------------
 *
 * Dump the geometrical model in a file.
 */

int
insituWriteGeom(InsituGeom *g, char *geomFile)
{
   char geomFullFile[PATH_MAX];
   InsituUpdateHooks *h;
   InsituPvFrame *f;
   InsituLink **l;
   InsituPkg *p;
   FILE *file;
   int i,j;

   if (geomFile[0] == '/')
      strcpy(geomFullFile, geomFile);
   else {
      char *t;

      if ((t = strrchr(g->file, '/')) == NULL)
	 strcpy(geomFullFile, geomFile);
      else {
	 *(++t) = 0;
	 sprintf(geomFullFile, "%s/%s", g->file, geomFile);
	 *t = '/';
      }
   }

   file = fopen(geomFullFile, "w");
   if (file == NULL) return errno;

   fprintf(file, "#\n# Robot model\n#\n\n");

   /* require directives */
   for(p=g->pkgs; p!=NULL; p=p->next)
      fprintf(file, "require %s;\n", p->name);
   fprintf(file, "\n");

   /* frames declaration */
   for(f=g->frames; f!=NULL; f=f->next)
      fprintf(file, "frame %s;\n", f->name);
   fprintf(file, "\n");
   fprintf(file, "main %s;\n\n",
	   insituGetFrameNameById(g->frames, g->mainFrame));

   /* update hooks */
   for(h=g->updateHooks; h!=NULL; h=h->next)
      fprintf(file, "update %s;\n", h->name);
   fprintf(file, "\n");

   /* links definition */
   l = g->graph->node;
   for(i=0;i<g->graph->nodes;i++) for(j=0;j<g->graph->nodes;j++, l++)
      if (*l != NULL) {
	 fprintf(file, "link {\n   from %s;\n   to %s;\n",
		 insituGetFrameNameById(g->frames, i),
		 insituGetFrameNameById(g->frames, j));
	 if ((*l)->type == ISLINK_STATIC) {
	    T3D t3d = (*l)->t3d;
	    t3dConvertTo(T3D_EULER, &t3d);
	    fprintf(file, "   translate\t%f %f %f;\n   rotate\t%f %f %f;\n",
		    t3d.euler.euler[3],
		    t3d.euler.euler[4],
		    t3d.euler.euler[5],
		    t3d.euler.euler[0],
		    t3d.euler.euler[1],
		    t3d.euler.euler[2]);
	 } else if ((*l)->update != NULL)
	    fprintf(file, "   update %s;\n", (*l)->updateName);

	 fprintf(file,"}\n\n");
      }

   fclose(file);
   return 0;
}


/*
 * insituDeleteGeom -----------------------------------------------------
 *
 * Destroys a robot description and frees any strorage allocated for it.
 */

void
insituDeleteGeom(InsituGeom *geom)
{
   InsituGeom *g;
   InsituPvFrame *f;

   if (--geom->refCount) return;

#ifdef VXWORKS
   if (geomSemId == NULL) { /* should never happen */
      fprintf(stderr, "insituDeleteRobot: NULL semaphore\n");
      return;
   }
   if (semTake(geomSemId, WAIT_FOREVER) != OK) return;
#endif

   /* Actually destroys model */
   if (geom == geoms)
      geoms = geom->next;
   else {
      for(g = geoms; g->next != geom && g != NULL; g = g->next)
	 /* empty body */;
      if (g != NULL) g->next = geom->next;
   }

   free(geom->path);
   free(geom->file);
   while((f = geom->frames) != NULL) {
      geom->frames = geom->frames->next;
      insituDeleteFrame(f);
   }
   if (geom->graph != NULL) insituDeleteGraph(geom->graph);
   free(geom);

#ifdef VXWORKS
   /* Destroy semaphore if there are no more models */
   if (geoms == NULL) {
      /* We must set the global semId to NULL before destroying it to
       * handle the case of a concurent semTake in progress */
      SEM_ID tmp = geomSemId;
      geomSemId = NULL;
      semDelete(tmp);
   } else
      semGive(geomSemId);
#endif
}


/*
 * newGeom --------------------------------------------------------------
 *
 * Allocate an InsituGeom structure and initialize it with reasonable
 * default values.
 */

static InsituGeom *
newGeom(ino_t ino, dev_t dev, char *path, char *file)
{
   InsituGeom *g;
   char *c;

   g = malloc(sizeof(InsituGeom));
   if (g == NULL) {
      errno = ENOMEM;
      return NULL;
   }

   g->refCount = 1;
   g->ino = ino;
   g->dev = dev;

   c = strrchr(file, '/');
   if (*c != '\0') {

      g->path = malloc((c-file+1) + 1);
      if (g->path == NULL) {
	 free(g);
	 errno = ENOMEM;
	 return NULL;
      }
      strncpy(g->path, file, c-file);
      /* Rich W. found this one... guys... what a shame! */
      g->path[c-file] = '\0';

   } else {
     g->path = malloc(strlen(path)+1);
      if (g->path == NULL) {
	 free(g);
	 errno = ENOMEM;
	 return NULL;
      }

      strcpy(g->path, path);
   }

   g->file = malloc(strlen(file)+1);
   if (g->file == NULL) {
      free(g->path);
      free(g);
      errno = ENOMEM;
      return NULL;
   }
   strcpy(g->file, file);

   g->pkgs = NULL;
   g->frames = NULL;
   g->mainFrame = -1;
   g->updateHooks = NULL;
   g->graph = NULL;

   return g;
}


/*
 * insituNewUpdateHook -------------------------------------------------
 *
 */

InsituUpdateHooks *
insituNewUpdateHook(char *name, InsituUpdateHook hook)
{
   InsituUpdateHooks *h;

   h = malloc(sizeof(InsituUpdateHooks));
   if (h == NULL) { errno = ENOMEM; return NULL; }

   h->name = malloc(strlen(name)+1);
   if (h->name != NULL) strcpy(h->name, name);
   h->hook = hook;
   h->next = NULL;
   return h;
}
