/*	$Id: geomparse.y 2007/08/21 18:55:07 mallet $ */

/*
 * Copyright (c) 2000-2007 LAAS/CNRS                  --  Tue May  9 2000
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

%{

#ifdef VXWORKS
#include <vxWorks.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <stdarg.h>
#include <string.h>
#include <limits.h>

#include "insitu.h"
#include "frame.h"
#include "geom.h"

extern int geomlex(/* YYSTYPE *lvalp */);

int geomLineNumber;
char geomFileName[PATH_MAX];

static void	yyerror(char *s);
static void	yylerror(int l, char *s, ...);
static void	yylwarning(int l, char *s, ...);

#define YYPARSE_PARAM geom

%}

%pure_parser

%union
{
   Y_UNION(int, ival);
   Y_UNION(double, dval);
   Y_UNION(char *, sval);
   Y_UNION(InsituPkg *, pkg);
   Y_UNION(InsituPvFrame *, frame);
   Y_UNION(InsituUpdateHooks *, update);
   Y_UNION(struct linkdecl, lnkdcl);
}

%token <ival> Y_REQUIRE
%token <ival> Y_FRAME Y_MAIN Y_LINK
%token <ival> Y_FROM Y_TO Y_TRANSLATE Y_ROTATE Y_UPDATE
%token <dval> Y_DOUBLE
%token <sval> Y_TOKEN

%type <ival> file decl
%type <sval> token
%type <pkg> require
%type <frame> frame framelist framedecl main
%type <update> update
%type <lnkdcl> link linkdecl linkelt

%start file

%%

file:
	decl
	{
	   $$.ival = 0;
	}
	| decl file
	{
	   $$.ival = 0;
	};

decl:
        require
        {
           if ($1.pkg == NULL) YYABORT;

           $1.pkg->next = ((InsituGeom *)geom)->pkgs;
           ((InsituGeom *)geom)->pkgs = $1.pkg;
           $$.ival = 0;
        }
	| frame
	{
	   if (((InsituGeom *)geom)->frames != NULL) {
	      InsituPvFrame *t;
	      for(t=((InsituGeom *)geom)->frames ;
		  t->next != NULL ; t = t->next) /* empty body */;
	      t->next = $1.frame;
	   } else
	      ((InsituGeom *)geom)->frames = $1.frame;
	   $$.ival = 0;
	}
	| main
	{
	   if ($1.frame != NULL)
	      ((InsituGeom *)geom)->mainFrame = insituGetFrameId($1.frame);
	   $$.ival = 0;
	}
	| update
	{
	   InsituUpdateHooks *h;

	   for(h=((InsituGeom *)geom)->updateHooks;h!=NULL;h=h->next)
	      if (h->hook == $1.update->hook) {
		 yylerror($1.l, "duplicate hook");
		 YYABORT;
	      }

	   $1.update->next = ((InsituGeom *)geom)->updateHooks;
	   ((InsituGeom *)geom)->updateHooks = $1.update;
	   $$.ival = 0;
	}
	| link
	{
	   int a, b;
	   T3D t3d;
	   InsituLinkType type;

	   a = insituGetFrameId($1.lnkdcl.from);
	   b = insituGetFrameId($1.lnkdcl.to);

	   if (!insituTestValidLink(((InsituGeom *)geom)->graph, a, b)) {
	      yylerror($1.l, "structural inconsistency for this link");
	      yylerror($1.l, "link may be declared twice");
	      yylerror($1.l, "or declared backward elsewhere");
	      YYABORT;
	   }

	   if (!($1.lnkdcl.flag & (ISLNKDCL_TRANSLATE|ISLNKDCL_ROTATE)))
	      type = ISLINK_DYNAMIC;

	   else
	      type = ISLINK_STATIC;

	   ((InsituGeom *)geom)->graph =
	      insituMakeLink(((InsituGeom *)geom)->graph, type, a, b);
	   t3d = t3dBryanIdentity;
	   t3d.euler.euler[0] = $1.lnkdcl.rotate[0];
	   t3d.euler.euler[1] = $1.lnkdcl.rotate[1];
	   t3d.euler.euler[2] = $1.lnkdcl.rotate[2];
	   t3d.euler.euler[3] = $1.lnkdcl.translate[0];
	   t3d.euler.euler[4] = $1.lnkdcl.translate[1];
	   t3d.euler.euler[5] = $1.lnkdcl.translate[2];
	   t3dConvertTo(T3D_MATRIX, &t3d);

	   insituSetLink(((InsituGeom *)geom)->graph, a, b, &t3d);
	   if (type == ISLINK_DYNAMIC) {
	      insituSetLinkUpdate(
		 ((InsituGeom *)geom)->graph, a, b,
		 $1.lnkdcl.updateName, $1.lnkdcl.update);
	      free($1.lnkdcl.updateName);
	   }

	   $$.ival = 0;
	};


/*
 * Require --------------------------------------------------------------
 */

require:
        Y_REQUIRE token ';'
        {
           $$.l = $1.l;
	   $$.pkg = insituLoadPackage(((InsituGeom *)geom)->path, $2.sval);
	   free($2.sval);
           if ($$.pkg == NULL) YYABORT;
        };


/*
 * Frame ----------------------------------------------------------------
 */

frame:
	Y_FRAME framelist ';'
	{
	   $$.l = $1.l;
	   $$.frame = $2.frame;
	};

framelist:
	framedecl
	{
	   $$ = $1;
	}
	| framedecl ',' framelist
	{
	   if (insituLookForFrame($1.frame->name, $3.frame) != NULL) {
	      yylwarning($1.l, "duplicate frame \"%s\"", $1.frame->name);
	      insituDeleteFrame($1.frame);
	      $$ = $3;
	   } else {
	      $1.frame->next = $3.frame;
	      $$ = $1;
	   }
	};

framedecl:
	token
	{
	   $$.l = $1.l;
	   $$.frame =
	      insituLookForFrame($1.sval, ((InsituGeom *)geom)->frames);
	   if ($$.frame != NULL) {
	      yylwarning($1.l, "duplicate frame \"%s\"", $1.sval);
	      $$.frame = NULL;
	   } else {
	      $$.frame = insituNewFrame($1.sval);
	      if ($$.frame == NULL) {
		 switch(errno) {
		    case EINVAL:
		       yylerror($1.l,
				"frame name too long for %s", $1.sval);
		       break;
		    case ENOMEM:
		       yylerror($1.l, "not enough memory");
		       break;
		   default:
		       yylerror($1.l, "unkown error");
		       break;
		 }
		 YYABORT;
	      }
	   }
	   free($1.sval);
	};


/*
 * Main -----------------------------------------------------------------
 */

main:
	Y_MAIN token ';'
	{
	   $$.l = $1.l;
	   $$.frame = insituLookForFrame(
	      $2.sval, ((InsituGeom *)geom)->frames);
	   if ($$.frame == NULL) {
	      yylerror($1.l, "unknown frame \"%s\"", $2.sval);
	      free($2.sval);
	      YYABORT;
	   }
	   free($2.sval);
	};


/*
 * Update ---------------------------------------------------------------
 */

update:
	Y_UPDATE token ';'
	{
	   InsituUpdateHook hook;
	   $$.l = $1.l;

	   hook = (InsituUpdateHook)
	      insituGetSymbol(((InsituGeom *)geom)->pkgs, $2.sval);
	   if (hook == NULL) {
	      yylerror($1.l, "undefined symbol \"%s\"", $2.sval);
	      free($2.sval);
	      YYABORT;
	   }

	   $$.update = insituNewUpdateHook($2.sval, hook);
	   free($2.sval);
	   if ($$.update == NULL) {
	      yylerror($1.l, "not enough memory");
	      YYABORT;
	   }
	};


/*
 * Link -----------------------------------------------------------------
 */

link:
	Y_LINK '{' linkdecl '}'
	{
	   $$ = $3;
	   if (!($$.lnkdcl.flag & (ISLNKDCL_FROM|ISLNKDCL_TO))) {
	      yylerror($1.l, "incomplete link description");
	      YYABORT;
	   }
	   if (!($$.lnkdcl.flag & (ISLNKDCL_TRANSLATE|ISLNKDCL_ROTATE)) &&
	       !($$.lnkdcl.flag & (ISLNKDCL_UPDATE))) {
	      yylerror($1.l, "incomplete link description");
	      YYABORT;
	   }
	   if (($$.lnkdcl.flag & (ISLNKDCL_TRANSLATE|ISLNKDCL_ROTATE)) &&
	       ($$.lnkdcl.flag & (ISLNKDCL_UPDATE))) {
	      yylerror($1.l, "link cannot be static and dynamic");
	      YYABORT;
	   }
	   if (!($$.lnkdcl.flag & ISLNKDCL_TRANSLATE)) {
	      $$.lnkdcl.translate[0] = 0.0;
	      $$.lnkdcl.translate[1] = 0.0;
	      $$.lnkdcl.translate[2] = 0.0;
	   }
	   if (!($$.lnkdcl.flag & ISLNKDCL_ROTATE)) {
	      $$.lnkdcl.rotate[0] = 0.0;
	      $$.lnkdcl.rotate[1] = 0.0;
	      $$.lnkdcl.rotate[2] = 0.0;
	   }
	};

linkdecl:
	linkelt ';'
	{
	   $$ = $1;
	}
	| linkelt ';' linkdecl
	{
	   $$ = $3;
	   if ($3.lnkdcl.flag & $1.lnkdcl.flag)
	      switch($1.lnkdcl.flag) {
		 case ISLNKDCL_FROM:
		    yylwarning($1.l, "duplicate \"from\" field");
		    break;
		 case ISLNKDCL_TO:
		    yylwarning($1.l, "duplicate \"to\" field");
		    break;
		 case ISLNKDCL_TRANSLATE:
		    yylwarning($1.l, "duplicate \"translate\" field");
		    break;
		 case ISLNKDCL_ROTATE:
		    yylwarning($1.l, "duplicate \"rotate\" field");
		    break;
		 case ISLNKDCL_UPDATE:
		    yylwarning($1.l, "duplicate \"update\" field");
		    break;
	      }
	   else {
	      $$.lnkdcl.flag |= $1.lnkdcl.flag;
	      switch($1.lnkdcl.flag) {
		 case ISLNKDCL_FROM:
		    $$.lnkdcl.from = $1.lnkdcl.from;
		    break;
		 case ISLNKDCL_TO:
		    $$.lnkdcl.to = $1.lnkdcl.to;
		    break;
		 case ISLNKDCL_TRANSLATE:
		    $$.lnkdcl.translate[0] = $1.lnkdcl.translate[0];
		    $$.lnkdcl.translate[1] = $1.lnkdcl.translate[1];
		    $$.lnkdcl.translate[2] = $1.lnkdcl.translate[2];
		    break;
		 case ISLNKDCL_ROTATE:
		    $$.lnkdcl.rotate[0] = $1.lnkdcl.rotate[0];
		    $$.lnkdcl.rotate[1] = $1.lnkdcl.rotate[1];
		    $$.lnkdcl.rotate[2] = $1.lnkdcl.rotate[2];
		    break;
		 case ISLNKDCL_UPDATE:
		    $$.lnkdcl.update = $1.lnkdcl.update;
		    $$.lnkdcl.updateName = $1.lnkdcl.updateName;
		    break;
	      }
	   }
	};

linkelt:
	Y_FROM token
	{
	   $$.l = $1.l;
	   $$.lnkdcl.flag = ISLNKDCL_FROM;
	   $$.lnkdcl.from = insituLookForFrame(
	      $2.sval, ((InsituGeom *)geom)->frames);
	   if ($$.lnkdcl.from == NULL) {
	      yylerror($1.l, "unknown frame \"%s\"", $2.sval);
	      free($2.sval);
	      YYABORT;
	   }
	   free($2.sval);
	}
	| Y_TO token
	{
	   $$.l = $1.l;
	   $$.lnkdcl.flag = ISLNKDCL_TO;
	   $$.lnkdcl.to = insituLookForFrame(
	      $2.sval, ((InsituGeom *)geom)->frames);
	   if ($$.lnkdcl.to == NULL) {
	      yylerror($1.l, "unknown frame \"%s\"", $2.sval);
	      free($2.sval);
	      YYABORT;
	   }
	   free($2.sval);
	}
	| Y_TRANSLATE Y_DOUBLE Y_DOUBLE Y_DOUBLE
	{
	   $$.l = $1.l;
	   $$.lnkdcl.flag = ISLNKDCL_TRANSLATE;
	   $$.lnkdcl.translate[0] = $2.dval;
	   $$.lnkdcl.translate[1] = $3.dval;
	   $$.lnkdcl.translate[2] = $4.dval;
	}
	| Y_ROTATE Y_DOUBLE Y_DOUBLE Y_DOUBLE
	{
	   $$.l = $1.l;
	   $$.lnkdcl.flag = ISLNKDCL_ROTATE;
	   $$.lnkdcl.rotate[0] = $2.dval;
	   $$.lnkdcl.rotate[1] = $3.dval;
	   $$.lnkdcl.rotate[2] = $4.dval;
	}
	| Y_UPDATE token
	{
	   $$.l = $1.l;
	   $$.lnkdcl.flag = ISLNKDCL_UPDATE;
           $$.lnkdcl.update = (InsituMatrix)
              insituGetSymbol(((InsituGeom *)geom)->pkgs, $2.sval);
           if ($$.lnkdcl.update == NULL) {
              yylerror($1.l, "unknown symbol \"%s\"", $2.sval);
              free($2.sval);
              YYABORT;
           }
	   $$.lnkdcl.updateName = $2.sval;
	};


/*
 * Token ----------------------------------------------------------------
 */

token:
	Y_TOKEN
	{
	   if ($1.sval == NULL) {
	      yylerror($1.l, "not enough memory");
	      YYABORT;
	   }
	   $$ = $1;
	}

%%

static void
yyerror(char *s)
{
   yylerror(geomLineNumber, s);
}

static void
yylerror(int l, char *s, ...)
{
   char *f;
   va_list list;

   if ((f = strrchr(geomFileName, '/')) == NULL)
      f = geomFileName;
   else f++;

   va_start(list, s);
   fprintf(stderr, "%s:%d: ", f, l);
   vfprintf(stderr, s, list);
   fprintf(stderr, "\n");
   va_end(list);
}

static void
yylwarning(int l, char *s, ...)
{
   char *f;
   va_list list;

   if ((f = strrchr(geomFileName, '/')) == NULL)
      f = geomFileName;
   else f++;

   va_start(list, s);
   fprintf(stderr, "%s:%d: warning: ", f, l);
   vfprintf(stderr, s, list);
   fprintf(stderr, "\n");
   va_end(list);
}


/*
 * insituCleanUpGeom ----------------------------------------------------
 *
 * Destroys any unsued stuff.
 */

int
insituCleanUpGeom(InsituGeom *g)
{
   InsituPvFrame *t, *u;

   if (g->mainFrame < 0) {
      yylerror(geomLineNumber, "no main frame defined");
      return -1;
   }

   while(g->frames != NULL && g->frames->id < 0) {
      yylwarning(geomLineNumber, "unused frame \"%s\"", g->frames->name);
      t = g->frames;
      g->frames = g->frames->next;
      insituDeleteFrame(t);
   }

   for(t = g->frames; t->next != NULL;)
      if (t->next->id < 0) {
	 yylwarning(geomLineNumber, "unused frame \"%s\"", t->next->name);
	 u = t->next;
	 t->next = t->next->next;
	 insituDeleteFrame(u);
      } else
	 t = t->next;

   return 0;
}
