/*	$Id: graph.c 2009/09/01 14:58:41 mallet $ */

/*
 * Copyright (c) 2000-2007,2009 LAAS/CNRS                  --  Mon May 15 2000
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
#ident "$Id: graph.c 2009/09/01 14:58:41 mallet $"

#ifdef VXWORKS
#include <vxWorks.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "insitu/lib.h"
#include "graph.h"


/* Local functions --------------------------------------------------- */

static InsituGraph *	newGraph(void);
static InsituGraph *	resizeGraph(InsituGraph *graph, int nodes);


/*
 * insituMakeLink --------------------------------------------------------
 *
 * Create a new link in graph "graph" between from the node a to the node
 * b. If the graph is too small (not enough nodes) it is enlarged.
 */

#define NODES_PACKET	15
#define gmax(a,b)	((a)>(b)?(a):(b))

InsituGraph *
insituMakeLink(InsituGraph *graph, InsituLinkType type, int a, int b)
{
   InsituLink *l;

   l = malloc(sizeof(InsituLink));
   if (l == NULL) {
      errno = ENOMEM;
      insituDeleteGraph(graph);
      return NULL;
   }
   l->type = type;
   l->update = NULL;
   l->updateName = NULL;

   if (graph == NULL) graph = newGraph();
   if (a >= graph->nodes || b >= graph->nodes) {
      graph =
	 resizeGraph(graph, gmax(gmax(a+1,b+1),graph->nodes + NODES_PACKET));
      if (graph == NULL) {
	errno = ENOMEM;
	return NULL;
      }
   }

   graph->node[a*graph->nodes+b] = l;
   return graph;
}

#undef gmax
#undef NODES_PACKET


/*
 * insituTestValidLink --------------------------------------------------
 *
 * Test if a link from a to b can be created. This function returns true if
 * the graph does neither contain a link from a to b nor from b to a.
 */

int
insituTestValidLink(InsituGraph *graph, int a, int b)
{
   if (graph == NULL) return 1;
   if (a >= graph->nodes || b >= graph->nodes) return 1;

   if (graph->node[a*graph->nodes+b] != NULL) return 0;
   if (graph->node[b*graph->nodes+a] != NULL) return 0;

   return 1;
}


/*
 * insituSetLink --------------------------------------------------------
 *
 * Defines the transformation for a link.
 */

void
insituSetLink(InsituGraph *graph, int a, int b, T3D *t3d)
{
   InsituLink *l = graph->node[a*graph->nodes+b];
   if (l == NULL) return;

   t3dCopyIn(&l->t3d, t3d);
   t3dInvert(t3dCopyIn(&l->d3t, t3d));

   t3dConvertTo(T3D_MATRIX, &l->t3d);
   t3dConvertTo(T3D_MATRIX, &l->d3t);
   l->t3d.flags = T3D_ALLOW_CONVERSION;
   l->d3t.flags = T3D_ALLOW_CONVERSION;
}

void
insituSetInvLink(InsituGraph *graph, int a, int b, T3D *t3d)
{
   InsituLink *l = graph->node[a*graph->nodes+b];
   if (l == NULL) return;

   t3dInvert(t3dCopyIn(&l->t3d, t3d));
   t3dCopyIn(&l->d3t, t3d);

   t3dConvertTo(T3D_MATRIX, &l->t3d);
   t3dConvertTo(T3D_MATRIX, &l->d3t);
   l->t3d.flags = T3D_ALLOW_CONVERSION;
   l->d3t.flags = T3D_ALLOW_CONVERSION;
}

void
insituSetLinkUpdate(InsituGraph *graph, int a, int b, char *name,
		    InsituMatrix update)
{
   InsituLink *l = graph->node[a*graph->nodes+b];
   if (l == NULL) return;

   l->update = update;
   if (l->updateName != NULL) free(l->updateName);
   l->updateName = malloc(strlen(name)+1);
   if (l->updateName != NULL) strcpy(l->updateName, name);
}


/*
 * insituUpdateGraph ----------------------------------------------------
 *
 * Invoke update function for every link.
 */

int
insituUpdateGraph(InsituGraph *graph)
{
   InsituLink **l;
   int i;

   l = graph->node;
   for(i=0; i<graph->nodes*graph->nodes; i++, l++) {
      if (*l == NULL) continue;
      if ((*l)->type == ISLINK_STATIC) continue;
      if ((*l)->update == NULL) continue;

      t3dCopyIn(&(*l)->t3d, ((*l)->update)());
      (*l)->t3d.flags = T3D_ALLOW_CONVERSION;
      t3dInvert(t3dCopyIn(&(*l)->d3t, &(*l)->t3d));

      t3dConvertTo(T3D_MATRIX, &(*l)->t3d);
      t3dConvertTo(T3D_MATRIX, &(*l)->d3t);
   }

   return 0;
}


/*
 * insituGetPath --------------------------------------------------------
 *
 * Computes and allocates a path that links the node a to the node b.
 */

static InsituPath *	getPath(InsituGraph *graph, InsituPath *start,
				int a, int b);
static InsituPath *	newPath(InsituLinkDir d, InsituLink *l,
				InsituPath *n);


InsituPath *
insituGetPath(InsituGraph *graph, int a, int b)
{
   if (a == b) return newPath(ISLINK_FORWARD, NULL, NULL);
   return getPath(graph, NULL, a, b);
}

static InsituPath *
getPath(InsituGraph *graph, InsituPath *start, int a, int b)
{
   int i;
   InsituPath p, *rest;
   InsituLink **l;

   /* if there's a forward solution, return it */
   l = graph->node + a*graph->nodes+b;
   if (*l != NULL) return newPath(ISLINK_FORWARD, *l, NULL);

   /* if there's a backward solution, return it */
   l = graph->node + b*graph->nodes+a;
   if (*l != NULL) return newPath(ISLINK_BACKWARD, *l, NULL);

   p.next = start;

   /* try every other nodes (forward) */
   l = graph->node + a*graph->nodes;
   for(i=0; i<graph->nodes; i++, l++) {
      if (*l == NULL) continue;

      /* avoid loops */
      for(rest = start; rest != NULL; rest = rest->next)
	 if (rest->link == *l) break;
      if (rest != NULL) continue;

      p.link = *l;
      rest = getPath(graph, &p, i, b);
      if (rest != NULL) return newPath(ISLINK_FORWARD, *l, rest);
   }

   /* try every other nodes (backward) */
   l = graph->node + a;
   for(i=0; i<graph->nodes; i++, l+=graph->nodes) {
      if (*l == NULL) continue;

      /* avoid loops */
      for(rest = start; rest != NULL; rest = rest->next)
	 if (rest->link == *l) break;
      if (rest != NULL) continue;

      p.link = *l;
      rest = getPath(graph, &p, i, b);
      if (rest != NULL) newPath(ISLINK_BACKWARD, *l, rest);
   }

   /* didn't find anything */
   return NULL;
}

static InsituPath *
newPath(InsituLinkDir d, InsituLink *l, InsituPath *n)
{
   InsituPath *p = malloc(sizeof(InsituPath));
   if (p == NULL) {
      errno = ENOMEM;
      return NULL;
   }

   p->dir = d;
   p->link = l;
   p->next = n;
   return p;
}


/*
 * insituDeletePath -----------------------------------------------------
 *
 * Delete a path
 */

void
insituDeletePath(InsituPath *path)
{
   InsituPath *p;

   while(path != NULL) {
      p = path;
      path = path->next;
      free(p);
   }
}

/*
 * newGraph -------------------------------------------------------------
 */

static InsituGraph *
newGraph()
{
   InsituGraph *graph;

   graph = malloc(sizeof(InsituGraph));
   if (graph == NULL) {
      errno = ENOMEM;
      return NULL;
   }

   graph->node = NULL;
   graph->nodes = 0;
   return graph;
}


/*
 * resizeGraph ----------------------------------------------------------
 */

static InsituGraph *
resizeGraph(InsituGraph *graph, int nodes)
{
   int i;
   InsituLink **l, **m;

   /* (re)allocate a (nodes x nodes) block of memory */
   if (graph->node == NULL)
      graph->node = malloc(nodes*nodes*sizeof(InsituLink *));
   else
      graph->node = realloc(graph->node, nodes*nodes*sizeof(InsituLink *));

   if (graph->node == NULL) {
      free(graph);
      errno = ENOMEM;
      return NULL;
   }

   /* in case of widening, move the existing nodes to the correct location and
    * zero the new links */
   if (graph->nodes < nodes) {
     /* move existing nodes */
     l = graph->node + graph->nodes * (graph->nodes - 1);
     m = graph->node + nodes * (graph->nodes - 1);
     for(i=graph->nodes-1; i>0; i--, l-=graph->nodes, m-=nodes)
       memmove(m, l, graph->nodes*sizeof(*l));

     /* zero new links */
     l = graph->node + graph->nodes;
     for(i=0; i<graph->nodes; i++) {
       memset(l, 0, (nodes-graph->nodes)*sizeof(InsituLink *));
       l += nodes;
     }
     l -= graph->nodes;
     memset(l, 0, nodes*(nodes-graph->nodes)*sizeof(InsituLink *));
   }

   graph->nodes = nodes;
   return graph;
}


/*
 * insituDeleteGraph ----------------------------------------------------
 */

void
insituDeleteGraph(InsituGraph *graph)
{
   InsituLink **l;
   int i;

   if (graph->node != NULL) {
      for(i=0, l=graph->node; i<graph->nodes*graph->nodes; i++, l++)
	 if (*l != NULL) free(*l);

      free(graph->node);
   }

   free(graph);
}
