/*	$Id: insitu.c 2009/09/01 15:22:47 mallet $ */

/*
 * Copyright (c) 2000-2007,2009 LAAS/CNRS                  --  Wed May 10 2000
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdlib.h>
#include <string.h>
#include <stddef.h>
#include <errno.h>

#include "insitu.h"


/*
 * insituClientCreate ---------------------------------------------------
 *
 * Opens a connection to the library and read a decsription file. The model
 * is open with permissions "mode" (either read only, read/write or
 * edit). Read only connections let the calling process get the current
 * frames configurations. Read/write connection let the caller modify
 * dynamic joints. Edit connections also let the caller modify the static
 * links.
 */

InsituModelId
insituOpen(char *geomFile, InsituAccesMode mode)
{
   InsituModelId id;
   InsituGeom *g;

   g = insituNewGeom(geomFile);
   if (g == NULL) return NULL;

   id = malloc(sizeof(struct InsituModel));
   if (id == NULL) {
      errno = ENOMEM;
      return NULL;
   }

   id->mode = mode;
   id->geom = g;
   id->lids = NULL;
   return id;
}


/*
 * insituUpdate ---------------------------------------------------------
 *
 * Update dynamic links in the graph.
 */

int
insituUpdate(InsituModelId id)
{
   InsituUpdateHooks *h;

   /* invoke hooks */
   for(h=id->geom->updateHooks; h!=NULL; h=h->next) {
      errno = (h->hook)();
      if (errno) return errno;
   }

   /* update graph */
   errno = insituUpdateGraph(id->geom->graph);
   if (errno) return errno;

   return 0;
}


/*
 * insituGetExportSize --------------------------------------------------
 *
 * return the size of the insituFrames
 */

size_t
insituGetExportSize(InsituModelId id)
{
   int size;
   InsituPvFrame *f;

   for(size=0, f=id->geom->frames; f!=NULL; size++, f=f->next);
   return (size*(sizeof(InsituFrame)));

}


/*
 * insituExport ---------------------------------------------------------
 *
 * Fill a frameset structure. Returns the size.
 */

int
insituExport(InsituModelId id, InsituFrameSet *set)
{
   int n = 0;
   InsituFrame *export = set->frame;
   InsituPvFrame *f;

   for(n=0,f=id->geom->frames; f!=NULL; f=f->next) {

      strcpy(export->name, f->name);
      n++;

      if (f->path == NULL) {
	 f->path =
	    insituGetPath(id->geom->graph, f->id, id->geom->mainFrame);
	 if (f->path == NULL) {
	   set->n = n;
	   return -1;
	 }
      }

      if (insituGet(f, &export->frameToMain)) {
	set->n = n;
	return -1;
      }

      export++;
   }

   set->n = n;
   return (size_t)((char *)export - (char *)set->frame) +
     offsetof(InsituFrameSet, frame);
}


/*
 * insituOpenLink -------------------------------------------------------
 *
 * Returns a link identifier. A link is a direct (forward or backward)
 * connection in the graph.
 */

InsituLinkId
insituOpenLink(InsituModelId id, char *from, char *to)
{
   InsituPvFrame *f, *t;
   InsituPath *path;
   InsituLinkId lid;

   f = insituLookForFrame(from, id->geom->frames);
   if (f == NULL) { errno = EINVAL; return NULL; }

   t = insituLookForFrame(to, id->geom->frames);
   if (t == NULL) { errno = EINVAL; return NULL; }

   if (f == t)  { errno = EINVAL; return NULL; }

   path = insituGetPath(id->geom->graph, f->id, t->id);
   if (path == NULL) { errno = EINVAL; return NULL; }

   if (path->next != NULL) {
      /* the connection is not direct */
      insituDeletePath(path);
      errno = EINVAL;
      return NULL;
   }

   lid = malloc(sizeof(struct InsituLinkId));
   if (lid == NULL) {
      insituDeletePath(path);
      errno = ENOMEM;
      return NULL;
   }

   lid->mode = id->mode;
   lid->path = path;
   lid->next = id->lids;
   id->lids = lid;
   return lid;
}


/*
 * insituOpenFrame ------------------------------------------------------
 *
 * Returns a frame identifier. A frame is a set of links that connects the
 * given frame to the main frame.
 */

InsituFrameId
insituOpenFrame(InsituModelId id, char *frame)
{
   InsituPvFrame *f;

   f = insituLookForFrame(frame, id->geom->frames);
   if (f == NULL) { errno = EINVAL; return NULL; }

   if (f->path == NULL) {
      f->path = insituGetPath(id->geom->graph, f->id, id->geom->mainFrame);
      if (f->path == NULL) { errno = EINVAL; return NULL; }
   }

   return f;
}


/*
 * insituSet ------------------------------------------------------------
 *
 * Set a link.
 */

int
insituSet(InsituLinkId lid, T3D *t3d)
{
   InsituLink *link = lid->path->link;

   if (lid->mode == ISMODE_READONLY) return errno = EACCES;
   if (link == NULL) return errno = EINVAL;
   if (link->type == ISLINK_STATIC && lid->mode != ISMODE_EDIT)
      return errno = EACCES;
   if (lid->path->next != NULL) return errno = EINVAL;

   if (lid->path->dir == ISLINK_FORWARD) {
     t3dCopyIn(&link->t3d, t3d); link->t3d.flags = T3D_ALLOW_CONVERSION;
     t3dCopyIn(&link->d3t, t3d); link->d3t.flags = T3D_ALLOW_CONVERSION;
     t3dInvert(&link->d3t);
   } else {
     t3dCopyIn(&link->d3t, t3d); link->d3t.flags = T3D_ALLOW_CONVERSION;
     t3dCopyIn(&link->t3d, t3d); link->t3d.flags = T3D_ALLOW_CONVERSION;
     t3dInvert(&link->t3d);
   }

   if (!t3dOk(&link->t3d) || !t3dOk(&link->d3t)) return errno = EINVAL;

   return 0;
}


/*
 * insituGet ------------------------------------------------------------
 *
 * Get a frame.
 */

int
insituGet(InsituFrameId fid, T3D *t3d)
{
  T3D r;
  InsituPath *path = fid->path;
  if (path == NULL) { return errno = EINVAL; }

  r = t3dMatrixIdentity;
  r.flags = T3D_ALLOW_CONVERSION;

  if (path->link == NULL) {
    t3dConvertTo(t3d->type, &r);
    t3dCopyIn(t3d, &r);
    return 0;
  }

  while(path!=NULL) {
    if (path->dir == ISLINK_FORWARD)
      t3dComp(&r, &path->link->t3d);
    else
      t3dComp(&r, &path->link->d3t);
    if (!t3dOk(&r)) return errno = EINVAL;

    path = path->next;
  }

  t3dConvertTo(t3d->type, &r);
  t3dCopyIn(t3d, &r);
  return 0;
}


/*
 * insituWrite ----------------------------------------------------------
 *
 */

int
insituWrite(InsituModelId id, char *file)
{
   return insituWriteGeom(id->geom, file);
}


/*
 * insituClose ----------------------------------------------------------
 *
 */

void
insituClose(InsituModelId id)
{
   struct InsituLinkId *lid;

   insituDeleteGeom(id->geom);

   while(id->lids != NULL) {
      lid = id->lids;
      id->lids = id->lids->next;

      insituDeletePath(lid->path);
      free(lid);
   }

   free(id);
}

